"""
This comes from the MXMRelations Product !!!
see http://www.zope.org/Members/maxm/productList/mxmRelations

This class implements the relations between python objects as a dictionary
with a list of objects.
if we have 4 objects 'obj1', 'obj2', 'obj3' and 'obj4', and we want
to relate obj1 with obj2 and obj3, what results is a structure like this:

self.relations[
    {
        'obj1':['obj2', 'obj3'],
        'obj2':['obj1'],
        'obj3':['obj1']
    }
]

So if you you look up:

self.relations['obj1']
>>> ['obj2', 'obj3']
self.relations['obj3']
>>> ['obj1']

Naturally you should use the "_get(relation)" method that does this for you, or returns an
empty list if the key doesn't exist.
"""

import zLOG
def log_deprecated(message,summary='Deprecation Warning',severity=zLOG.WARNING):
    zLOG.LOG('CMF Relation Instances: ',severity,summary,message)

#generic log method
def log(message,summary='',severity=0):
    zLOG.LOG('CMF Relation Instances: ',severity,summary,message)
    


class RelationInstances:

    ########################################################
    # Helper methods

    relationInstances = {}

    def _toList(self, arg):
        "make the argument into a list if it isn't allready"
        if type(arg) != type([]):
            arg = [arg]
        return arg


    def _appendCreate(self, rel1, rel2):
        "Appends RelationInstances, or create the list with the first relation"
        log('rel1: %s' % str(rel1))
        log('rel2: %s' % str(rel2))
        log('known instances: %s, %s' % (self.relationInstances, len(self.relationInstances)))
        if len(self.relationInstances) > 0 :
            if self.relationInstances.has_key(rel1):
                # Duplicates are not allowed
                log('Key exists')
                if rel2 not in self.relationInstances[rel1]: 
                    self.relationInstances[rel1].append(rel2)
        else:
            # New key
            log('New key')
            self.relationInstances[rel1] = [rel2]


    def _unrelDelete(self, r1, r2):
        "Unrelates two RelationInstances and deletes the list if it's empty"
        if r2 in self.relationInstances[r1]:
            self.relationInstances[r1].remove(r2)
            if len(self.relationInstances[r1]) == 0:
                del(self.relationInstances[r1])




    def _relate(self, rel1, rel2):
        """
        Sets RelationInstances. If there allready is a key it appende the RelationInstances
        else it creates a new key with a list of RelationInstances
        """
        rel1 = self._toList(rel1)
        rel2 = self._toList(rel2)
        for r1 in rel1:
            for r2 in rel2:
                if r1 != r2:
                    self._appendCreate(r1, r2)
                    # self._appendCreate(r2, r1) Inverse ! we don't want that yet


    def _unrelate(self, rel1, rel2):
        """
        Removes RelationInstances from object rel1 to object rel2
        """
        rel1 = self._toList(rel1)
        rel2 = self._toList(rel2)
        for r1 in rel1:
            for r2 in rel2:
                self._unrelDelete(r1, r2)
                # self._unrelDelete(r2,r1) Inverse ! we don't want that yet


    def _delete(self, object):
        """
        Removes all references to and from the object.
        Used ie. if an object is deleted.
        """
        # Delete all refrences from this object to others
        if object in self.relationInstances.keys():
            del(self.relationInstances[object])
        # Delete all refs from others to this object
        for i in self.relationInstances.keys():
            self.relationInstances[i].remove(object)
            if len(self.relationInstances[i]) == 0:
                del(self.relationInstances[i]) # delete RelationInstances with empty lists
                
    

    def _get(self, rel):
        "Returns all RelationInstances of object <param1>, or an empty list"
        if len(self.relationInstances):
                if self.relationInstances.has_key(rel):
                    if len(self.relationInstances[rel]) > 0:
                        return self.relationInstances[rel]
                    else:
                        del(self.relationInstances[rel]) # delete RelationInstances with empty lists
                        return []
        else:
            return []


class A:
    pass

if __name__ == '__main__':
    # Test case

    a=A()
    b=A()
    rl = RelationInstances()
    rl._relate([a,b],[3,4,5])
    print rl.relationInstances
    rl._delete(3)
    print "Relations of 1:"
    print rl._get(a)
    rl._delete(4)
    print rl.relationInstances

##    rl = Relations()
##    rl._relate([1,2],[3,4,5])
##    rl._relate([1,2],[3,4,5])
##    print rl.relations
##    print rl._get(5)
##    rl._unrelate([1,2], 5)
##    print rl.relations
