from Products.CMFDefault.PropertiesTool import PropertiesTool as BaseTool
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from OFS.Folder import Folder
from Globals import InitializeClass, DTMLFile
from Products.CMFCore.utils import getToolByName
from RelationInstances import RelationInstances

#from Products.CMFCore.utils Relation
from OFS.PropertyManager import PropertyManager
from OFS.SimpleItem import SimpleItem
from string import join
from urllib import unquote
from AccessControl import ClassSecurityInfo
import urlparse
import re



import zLOG
def log_deprecated(message,summary='Deprecation Warning',severity=zLOG.WARNING):
    zLOG.LOG('CMF Relations: ',severity,summary,message)

#generic log method
def log(message,summary='',severity=0):
    zLOG.LOG('CMF Relations: ',severity,summary,message)
    
class RelationsTool(Folder, BaseTool):
    """ Specialized PropertiesTool that contains typed relations between objects """
    id = 'portal_relations'

    meta_type = 'Relations Tool'
    icon = 'tool.gif'
    meta_types = all_meta_types =  ( ( { 'name' : 'Relation'
                                       , 'action' : 'manage_addRelationForm' }, ) )

    manage_options = ( (Folder.manage_options[0], ) +
                        BaseTool.manage_options  )

    manage_addRelationForm = PageTemplateFile( 'www/addRelation', globals() )

    
    def title(self):
        """ return BaseTool title """
        return self.meta_type

    def addRelation(self, id, title='', propertysheet=None):
        """ add a new Relation """
        o=Relation(id, title)
        o._setId(id)
        self._setObject(id, o)
        o=self._getOb(id)
        o.title=title

        # copy the propertysheet values onto the new instance
        if propertysheet is not None:
            if not hasattr(propertysheet, 'propertyIds'):
                raise TypeError, 'propertysheet needs to be a PropertyManager'
            for property in propertysheet.propertyMap():
                pid=property.get('id')
                ptype=property.get('type')
                pvalue=propertysheet.getProperty(pid)
                o._setProperty(pid, pvalue, ptype)
            

    def unrelateObject(self, object):
        """ Deletes all relations to and from the object """
        # log('Unrelate Object called')
        for i in self.objectItems():
             i[1].unrelateObject(object)
        
             
    def unrelateFromObject(self, object, predicate=None):
        """ Deletes all relations from the object """
        # log('Unrelated Object called')
        if predicate is None: 
            for i in self.objectItems():
                 i[1].unrelateFromObject(object)
        else:
            for i in self.objectItems():
                if i[0] == predicate:
                        i[1].unrelateFromObject(object)
                 
    def unrelateToObject(self, object, predicate=None):
        """ Deletes all relations to the object """
        # log('Unrelated Object called')
        if predicate is None:
            for i in self.objectItems():
                 i[1].unrelateToObject(object)
        else:
            for i in self.objectItems():
                if i[0] == predicate:
                        i[1].unrelateToObject(object)

    
    def updateRename(self, folder, old_ids, new_ids):
        """ Updates all object links from their old_ids to the new_ids. """
        for i in self.objectItems():
            i[1].updateRename(folder, old_ids, new_ids)
                        
    def relate(self, subject, predicate, object):
        """ calls relate on relation predicate """
        # # log('relate - s:%s p:%s o:%s' % (subject,predicate, object))
        for i in self.objectItems():
            if i[0] == predicate:
                for n in object:
                    i[1].relate(subject,n)
                
    def manage_addRelation(self, id, title='', propertysheet=None, REQUEST=None):
        """ adds a instance of a relation 
            if handed a propertysheet put the
            properties into new propertysheet.
        """
##        if id is not None:
##            self.addRelation(id, title, propertysheet)
##        else:
##            return self.manage_addRelationForm
        self.addRelation(id, title, propertysheet)
        
        if REQUEST is not None:
            return self.manage_main()

    def getRelationsWithDomain(self, domain):
        """
        Returns a list of relation objects, which have the
        given domain.
        """
        ret = []
        for i in self.objectValues():
            if i.isCompatibleWithDomain(domain):
                ret.append(i)
        return ret

    def getRelationsWithRange(self, range):
        """
        Returns a list of relation objects, which have the
        given range.
        """
        ret = []
        for i in self.objectValues():
            if i.isCompatibleWithRange(range):
                ret.append(i)
        return ret

    def getAllRelationsFromObject(self, object):
        """
        Returns all relation instances for the object. It
        is  a dictionary with a list of objects for each
        relation.
            {
                'rel1':['obj2', 'obj3'],
                'rel2':['obj1'],
                'rel3':['obj1']
            }
        """
        ret = {}
        for i in self.objectValues():
            m = i.getRelationInstances(object)
            if len(m) > 0:
                ret[i] = m
        return ret

InitializeClass(RelationsTool)

class Struct:
    __allow_access_to_unprotected_subobjects__=1
    def __init__(self, **entries): self.__dict__.update(entries)



class Relation (PropertyManager, SimpleItem):    
    """
    A common base class for objects with configurable
    properties in a fixed schema.
    """

    relationInstances = {}

    ########################################################
    # Helper methods

    def __init__(self, id, title):
        # # log('Init called')
        self.relationInstances = {}
        self.id = id
        self.title = title
        
    def _toList(self, arg):
        "make the argument into a list if it isn't allready"
        if type(arg) != type([]):
            arg = [arg]
        return arg
    
    
    def updateRename(self, folder, old_ids, new_ids):
        """ Updates all object links from their old_ids to the new_ids. """
        folderpath = self._o2path(folder)
        # # log('UpdateRename %s all: %s new: %s ' % (folderpath, old_ids, new_ids))
        for i in range(len(old_ids)):
            oldpath = folderpath + '/' + old_ids[i]
            # # log(oldpath)
            newpath = folderpath + '/' + new_ids[i]
            # # log(newpath)
            if self.relationInstances.has_key(oldpath) :
                self.relationInstances[newpath] = self.relationInstances.get(oldpath)
                # # log(self.relationInstances)    
                # del self.relationInstances[oldpath]
            for k in self.relationInstances.keys():
                if oldpath in self.relationInstances.get(k):
                    self.relationInstances[k].append(newpath)
                    # self.relationInstances[k].remove(oldpath)
                    # # log(self.relationInstances)
        
    def _o2path(self, object):
        "Returns a paths as a string"
        # # log('object>path')
        try:
            return join(object.getPhysicalPath()[1:], '/')
        except Exception, e:
            # log(e)
            return object
    
    def _objectsToPaths(self, objList):
        """
        Takes a list of objects and returns a list of paths as strings
        of the form ['path/to/object/one','path/to/object/two']
        """
        paths = []
        for obj in objList:
            paths.append(self._o2path(obj))
        return paths


    def pathToObject(self, path):
        """
        returns a list of objects from a list of paths
        deletes any paths that are no longer valid because they have been
        deleted, moved, or renamed from Zope.
        """
        try:
            # get the object from the path
            path = unquote(path)
            return self.getPhysicalRoot().restrictedTraverse(path)
        except KeyError:
            # The object the path points to is deleted, moved,
            # or renamed from Zope.
            # So the relation is no longe valid. Delete it!
            self._delete(path)
        except 'Unauthorized':
            # If the user don't have access don't add the related object,
            # but don't delete the relation either
            pass



    def _appendCreate(self, rel1, rel2):
        "Appends RelationInstances, or create the list with the first relation"
        
        rel1 = self._o2path(rel1)
        rel2 = self._o2path(rel2)
        # log('rel1: %s' % str(rel1))
        # log('rel2: %s' % str(rel2))
        # log('known instances: %s, %s' % (str(self.relationInstances), len(self.relationInstances)))
        if len(self.relationInstances) > 0 :
            if self.relationInstances.has_key(rel1):
                # Duplicates are not allowed
                # # log('Key exists')
                if rel2 not in self.relationInstances[rel1]: 
                    self.relationInstances[rel1].append(rel2)
        else:
            # New key
            # # log('New key')
            self.relationInstances[rel1] = [rel2]
            


    def _unrelDelete(self, r1, r2):
        "Unrelates two RelationInstances and deletes the list if it's empty"
        if r2 in self.relationInstances[r1]:
            self.relationInstances[r1].remove(r2)
            if len(self.relationInstances[r1]) == 0:
                del self.relationInstances[r1] 




    def _relate(self, rel1, rel2):
        """
        Sets RelationInstances. If there allready is a key it appende the RelationInstances
        else it creates a new key with a list of RelationInstances
        """
        # log('Relate calles')
                    # self._appendCreate(r2, r1) Inverse ! we don't want that yet


    def _unrelate(self, rel1, rel2):
        """
        Removes RelationInstances from object rel1 to object rel2
        """
        rel1 = self._toList(rel1)
        rel2 = self._toList(rel2)
        for r1 in rel1:
            for r2 in rel2:
                self._unrelDelete(r1, r2)
                # self._unrelDelete(r2,r1) Inverse ! we don't want that yet


    def _delete(self, object):
        """
        Removes all references to and from the object.
        Used ie. if an object is deleted.
        """
        self._deleteFrom(object)
        self._deleteTo(object)
        
    def _deleteFrom(self, object):
        """ Delete all refrences from this object to others """
        if self.relationInstances.has_key(object):
            # # log('Key exists')
            del self.relationInstances[object]
            
    def _deleteTo(self, object):
        """ Delete all refs from others to this object """
        for i in self.relationInstances.keys():
            # # log('delete to in %s' % i)
            self.relationInstances[i].remove(object)
            if len(self.relationInstances[i]) == 0:
                del self.relationInstances[i]  # delete RelationInstances with empty lists
                
    

    def _get(self, rel):
        "Returns all RelationInstances of object <param1>, or an empty list"
        rel = self._o2path(rel)
        # # log('_get %s' % rel)
        if len(self.relationInstances) > 0:
            # # log('_get existing relation Instances: %s' % self.relationInstances)
            if self.relationInstances.has_key(rel):
                if len(self.relationInstances[rel]) > 0:
                   # # log(self.relationInstances[rel])
                   return self.relationInstances[rel]
                else:
                    del(self.relationInstances[rel]) # delete RelationInstances with empty lists
                    # # log('No entry for %s' % rel)
                    return []
        else:
            # # log('No entries at all')
            return []
    
    
    ### Zope Stuff
    
    my_manage_options=(
        {'label':'Properties', 'action':'manage_relationForm',},
        {'label':'Relation Instances', 'action':'manage_relationInstances', }
        )
    
    manage_relationForm = DTMLFile('www/manageRelation', globals(),
                                   property_extensible_schema__=1)
    
    manage_relationInstances = PageTemplateFile( 'www/viewRelationInstances', globals() )

    
    manage_options = ( my_manage_options
                     + SimpleItem.manage_options)

    _properties = (
                  {'id':'domain','type':'selection','mode':'w','select_variable':'listTypes'},
                  {'id':'range','type':'selection','mode':'w','select_variable':'listTypes'},
                  {'id':'uri','type':'string','mode':'w'},
                  {'id':'documentation','type':'text','mode':'w'},
                  )

    domain = range = uri = documentation = ''

    my_id=''

    def _setId(self,id):
        """Sets the id of this relation"""
        self.my_id = id
        #  # log('setId called with value %s' % id)

    def getId(self):
        """ Retrieves the id of this relation"""
        # # log('getId called, returning %s' % self.my_id)
        return self.my_id
    
    def listTypes(self) :
        """ Lists all Types that are available in the portal """
        types_tool = getToolByName(self, 'portal_types')
        return [ str(i) for i in types_tool.listContentTypes() ]

    def isCompatibleWithDomain(self, className):
        """ Checks, whether this relation is compatible with the given class <className> """
        return ( className == self.domain )

    def isCompatibleWithRange(self, className):
        """ Checks, whether this relation is compatible with the given class <className> """
        return ( className == self.range )


    def getPossibleValues(self):
        """ Returns a triples (object, uri, title) of all possible instances 
            Instances are retrieved using the portal_catalog
        """
        catalog = getToolByName(self, 'portal_catalog')
        # !!! TO DO inheritance siehe Zope Book ODER machen
        results=catalog( meta_type = self.range )
        return [ (result, result.getURL(), result.Title ) for result in results ]

    def relate(self, objectFrom, objectTo):
        """ Creates a relation from object <objectFrom> to object <objectTo> """
        # log('relate From:%s' % objectFrom)
        # log('relate To:%s' % objectTo)
        # log(type(str(objectTo)))
        if str(type(objectFrom)) == '<type \'string\'>':
            objectFrom = self._pathToObject(objectFrom)
        if str(type(objectTo)) == '<type \'string\'>':
            # log('true')
            objectTo = (self._pathToObject(objectTo))
            
        if not (objectTo is None or objectFrom is None):
            self._appendCreate(objectFrom, objectTo)
            self._p_changed = 1


    def _pathToObject(self, string):
        """ Internal function that maps path strings to objects """
        # pattern = re.compile(r'/')
        path = urlparse.urlparse(string)[2]
        try:
            obj = self.getPhysicalRoot().restrictedTraverse(path)
        except:
            # log('Object %s moved, deleted or access restriction not met' % string)
            obj= None
        # # log('path: %s -> object: %s' % (string,str(obj)))
        return obj


    def unrelateObject(self, object):
        """ Removes all relations to and from this object """
        path = self._o2path(object)
        self._delete(path)
        self._p_changed = 1
        
        
    def unrelateFromObject(self, object):
        """ Removes all relations from this object """
        path = self._o2path(object)        
        self._deleteFrom(path)
        self._p_changed = 1
        
    def unrelateToObject(self, object):
        """ Removes all relations to this object """
        path = self._o2path(object)
        self._deleteTo(path)
        self._p_changed = 1

    def getRelationInstances(self, object):
        """ Returns all objects that are related with the given object """
        x = [ self.pathToObject(i) for i in self._get(object) ]
        #log(x)
        #log(type(x))
        return x 
        
    def getAllRelationInstances(self):
        """ Returns all relation instances. This is a dictionary
            with a list of objects, e.g.
            {
                'obj1':['obj2', 'obj3'],
                'obj2':['obj1'],
                'obj3':['obj1']
            }
        """
        return self.relationInstances

    def getSmartList(self, object):
        """ Returns a list of objects with parameters (object, title, url, selected)
        which are related from "object".
        Very usable for making selection widgets. """
        result = []
        object = self._o2path(object)
        relationInstances = self.getRelationInstances(object)
        # # log('Relation Instances: %s' % relationInstances)
        for o in self.getPossibleValues():
            selected = 0
            if relationInstances is not None: 
                selected = (o[1] in [ i.absolute_url() for i in relationInstances ])
            result.append(
                Struct(
                    selected    = selected,
                    path        = o[1],
                    title       = o[2],
                    object      = o[0]
                    )
                )
        return result
        
    ## TO DO Remove me !!!
    #def index_html( self ):
        #""" For testing only """
        #ret = "<html><body>Values of Range %s:<p>" % self.range
        #for result in self.getPossibleValues() :
            #ret += "<a href=\"%s\">%s</a><br/>" % (result[0], result[1])
        #ret += "</body></html>"
        #return ret
        
InitializeClass( Relation )
